# w.c.s. - web application for online forms
# Copyright (C) 2005-2010  Entr'ouvert
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

import time
import re
import xml.etree.ElementTree as ET

from quixote import get_request, get_publisher
from quixote.html import htmltext, TemplateIO

from qommon.form import *
from qommon.misc import localstrftime, date_format, ellipsize
from qommon import get_cfg, get_logger

from qommon.strftime import strftime

import data_sources


def fixcsv(self, v):
    if type(s) is not str: return s
    if not s: return s
    return s


class PrefillSelectionWidget(CompositeWidget):
    def __init__(self, name, value = None, **kwargs):
        CompositeWidget.__init__(self, name, value, **kwargs)

        if not value:
            value = {}

        options = [('none', _('None')),
                ('string', _('String')),
                ('formula', _('Formula (Python)')),
                ('user', _('User Field'))]

        self.add(SingleSelectWidget, 'type', options = options, value = value.get('type'))

        self.parse()
        if not self.value:
            self.value = {}

        if self.value.get('type') in ('string', 'formula'):
            self.add(StringWidget, 'value', value = value.get('value'), size=80)
        elif self.value.get('type') == 'user':
            formdef = get_publisher().user_class.get_formdef()
            users_cfg = get_cfg('users', {})
            if formdef:
                user_fields = []
                for field in formdef.fields:
                    if field.label in [x[1] for x in user_fields]:
                        # do not allow duplicated field names
                        continue
                    user_fields.append((field.id, field.label))
                if not users_cfg.get('field_email'):
                    user_fields.append(('email', _('Email (builtin)')))
            else:
                user_fields = [('name', _('Name')), ('email', _('Email'))]

            self.add(SingleSelectWidget, 'value', value = value.get('value'),
                    options = user_fields)

        self.add(SubmitWidget, 'apply', value = _('Apply'))

        self._parsed = False


    def _parse(self, request):
        values = {}
        for name in ('type', 'value'):
            value = self.get(name)
            if value:
                values[name] = value
        self.value = values or None
        if values and values['type'] == 'formula' and values.get('value'):
            try:
                compile(values.get('value', ''), '<string>', 'eval')
            except (SyntaxError, TypeError), e:
                self.set_error(_('invalid expression: %s') % e)

    def render_content(self):
        r = TemplateIO(html=True)
        for widget in self.get_widgets():
            r += widget.render_content()
        return r.getvalue()


class Field:
    id = None
    type = None
    varname = None
    label = None
    extra_css_class = None
    convert_value_from_str = None
    convert_value_to_str = None
    in_listing = False
    prefill = None
    store_display_value = None

    anonymise = True
    stats = None

    def __init__(self, **kwargs):
        for k, v in kwargs.items():
            setattr(self, k.replace('-', '_'), v)

    def init(cls):
        pass
    init = classmethod(init)

    def get_admin_attributes(self):
        return ['label', 'type']

    def export_to_json(self, include_id=False):
        field = {}
        if include_id:
            extra_fields = ['id']
        else:
            extra_fields = []
        for attribute in self.get_admin_attributes() + extra_fields:
            if attribute in ('in_listing', ):
                continue
            if hasattr(self, attribute) and getattr(self, attribute) is not None:
                val = getattr(self, attribute)
                field[attribute] = val
        return field

    def init_with_json(self, elem, include_id=False):
        if include_id:
            self.id = elem.get('id')
        for attribute in self.get_admin_attributes():
            if attribute in elem:
                setattr(self, attribute, elem.get(attribute))

    def export_to_xml(self, charset, include_id=False):
        field = ET.Element('field')
        if include_id:
            extra_fields = ['id']
        else:
            extra_fields = []
        for attribute in self.get_admin_attributes() + extra_fields:
            if hasattr(self, attribute) and getattr(self, attribute) is not None:
                val = getattr(self, attribute)
                if type(val) is dict and not val:
                    continue
                el = ET.SubElement(field, attribute)
                if type(val) is dict:
                    for k, v in val.items():
                        ET.SubElement(el, k).text = unicode(v, charset, 'replace')
                elif type(val) is list:
                    if attribute[-1] == 's':
                        atname = attribute[:-1]
                    else:
                        atname = 'item'
                    for v in val:
                        ET.SubElement(el, atname).text = unicode(v, charset, 'replace')
                elif type(val) in (str, unicode):
                    if type(val) is unicode:
                        el.text = val
                    else:
                        el.text = unicode(val, charset, 'replace')
                else:
                    el.text = str(val)
        return field

    def init_with_xml(self, elem, charset, include_id=False):
        for attribute in self.get_admin_attributes():
            el = elem.find(attribute)
            if el is None:
                continue
            if el.getchildren():
                if type(getattr(self, attribute)) is list:
                    v = [x.text.encode(charset) for x in el.getchildren()]
                elif type(getattr(self, attribute)) is dict:
                    v = {}
                    for e in el.getchildren():
                        v[e.tag] = e.text.encode(charset)
                else:
                    print 'currently:', self.__dict__
                    print '  attribute:', attribute
                    # ???
                    raise AssertionError
                setattr(self, attribute, v)
            else:
                if el.text is None:
                    setattr(self, attribute, None)
                elif el.text in ('False', 'True'): # bools
                    setattr(self, attribute, eval(el.text))
                elif type(getattr(self, attribute)) is int:
                    setattr(self, attribute, int(el.text.encode(charset)))
                else:
                    setattr(self, attribute, el.text.encode(charset))
        if include_id:
            try:
                self.id = elem.find('id').text.encode(charset)
            except:
                pass

    def get_rst_view_value(self, value, indent=''):
        return indent + self.get_view_value(value)

    def get_csv_heading(self):
        return []

    def get_csv_value(self, element, hint=None):
        return []

    def get_prefill_value(self):
        t = self.prefill.get('type')
        if t == 'string':
            return self.prefill.get('value')

        elif t == 'user' and get_request().user:
            x = self.prefill.get('value')
            user = get_request().user
            if x == 'email':
                return user.email
            elif user.form_data:
                userform = user.get_formdef()
                for userfield in userform.fields:
                    if userfield.id == x:
                        return user.form_data.get(x)

        elif t == 'formula':
            formula = self.prefill.get('value')
            try:
                ret = eval(formula,
                    get_publisher().get_global_eval_dict(),
                    get_publisher().substitutions.get_context_variables())
                if ret:
                    return str(ret)
            except:
                pass

        return None

    def feed_session(self, value, display_value):
        pass


class WidgetField(Field):
    hint = None
    required = True
    in_listing = True
    extra_attributes = []
    prefill = {}

    def add_to_form(self, form, value = None):
        kwargs = {'required': self.required}
        if value:
            kwargs['value'] = value
        for k in self.extra_attributes:
            if hasattr(self, k):
                kwargs[k] = getattr(self, k)
        self.perform_more_widget_changes(form, kwargs)
        if self.hint and self.hint.startswith('<'):
            hint = htmltext(self.hint)
        else:
            hint = self.hint
        form.add(self.widget_class, 'f%s' % self.id, title=self.label, hint=hint, **kwargs)
        widget = form.get_widget('f%s' % self.id)
        widget.field = self
        if self.extra_css_class:
            if hasattr(widget, 'extra_css_class') and widget.extra_css_class:
                widget.extra_css_class = '%s %s' % (widget.extra_css_class, self.extra_css_class)
            else:
                widget.extra_css_class = self.extra_css_class
        if self.varname:
            widget.div_id = 'var_%s' % self.varname

    def perform_more_widget_changes(self, form, kwargs, edit = True):
        pass


    def add_to_view_form(self, form, value = None):
        kwargs = {}

        self.field_key = 'f%s' % self.id
        self.perform_more_widget_changes(form, kwargs, False)

        for k in self.extra_attributes:
            if hasattr(self, k):
                kwargs[k] = getattr(self, k)

        if self.widget_class is StringWidget and not 'size' in kwargs and value:
            # set a size if there is not one already defined, this will be for
            # example the case with ItemField
            kwargs['size'] = len(value)

        form.add(self.widget_class, self.field_key, title = self.label,
                value = value, readonly = 'readonly', **kwargs)

    def fill_admin_form(self, form):
        form.add(StringWidget, 'label', title = _('Label'), value = self.label,
                required = True, size = 50)
        form.add(CheckboxWidget, 'required', title = _('Required'),
                value = self.required)
        form.add(VarnameWidget, 'varname', title=_('Variable Name'),
                value=self.varname, size=30)
        form.add(TextWidget, 'hint', title = _('Hint'), value = self.hint,
                cols=60, rows=3)
        form.add(CheckboxWidget, 'in_listing', title = _('Display in listings'),
                value = self.in_listing)
        form.add(StringWidget, 'extra_css_class', title = _('Extra class for CSS styling'),
                value = self.extra_css_class, size = 30)
        form.add(PrefillSelectionWidget, 'prefill', title = _('Prefill'),
                value = self.prefill)

    def check_admin_form(self, form):
        return

    def get_admin_attributes(self):
        return Field.get_admin_attributes(self) + ['required', 'hint',
                'varname', 'in_listing',
                'extra_css_class', 'prefill']

    def get_csv_heading(self):
        return [self.label]

    def get_view_value(self, value):
        return str(value)

    def get_view_short_value(self, value, max_len = 30):
        return self.get_view_value(value)

    def get_csv_value(self, element, hint=None):
        if self.convert_value_to_str:
            return [self.convert_value_to_str(element)]
        return [element]

field_classes = []
field_types = []

def register_field_class(klass):
    if not klass in field_classes:
        field_classes.append(klass)
        if not issubclass(klass, WidgetField):
            field_types.append((klass.key, klass.description))
        else:
            non_widgets = [x for x in field_classes if not issubclass(x, WidgetField)]
            if not non_widgets:
                field_types.append((klass.key, klass.description))
            else:
                idx = field_types.index([x for x in field_types if x[0] == non_widgets[0].key][0])
                field_types.insert(idx, (klass.key, klass.description))
        klass.init()


class TitleField(Field):
    key = 'title'
    description = N_('Title')

    def add_to_form(self, form, value = None):
        if self.extra_css_class:
            extra_css_class = ' class="%s"' % self.extra_css_class
        else:
            extra_css_class = ''
        form.widgets.append(HtmlWidget(
                    htmltext('<h3%s>%s</h3>' % (extra_css_class, self.label))))
    add_to_view_form = add_to_form

    def fill_admin_form(self, form):
        form.add(StringWidget, 'label', title = _('Label'), value = self.label,
                required = True, size = 50)
        form.add(StringWidget, 'extra_css_class', title = _('Extra class for CSS styling'),
                value = self.extra_css_class, size = 30)

    def get_admin_attributes(self):
        return Field.get_admin_attributes(self) + ['extra_css_class']

register_field_class(TitleField)


class SubtitleField(TitleField):
    key = 'subtitle'
    description = N_('Subtitle')

    def add_to_form(self, form, value = None):
        if self.extra_css_class:
            extra_css_class = ' class="%s"' % self.extra_css_class
        else:
            extra_css_class = ''
        form.widgets.append(HtmlWidget(
                    htmltext('<h4%s>%s</h4>' % (extra_css_class, self.label))))
    add_to_view_form = add_to_form

register_field_class(SubtitleField)


class CommentField(Field):
    key = 'comment'
    description = N_('Comment')

    def add_to_form(self, form, value = None):
        if self.extra_css_class:
            extra_css_class = ' class="%s"' % self.extra_css_class
        else:
            extra_css_class = ''

        if '\n\n' in self.label:
            label = '<p>' + re.sub('\n\n+', '</p>\n<p>', self.label) + '</p>'
        else:
            label = self.label

        if '<p' in label:
            enclosing_tag = 'div'
        else:
            enclosing_tag = 'p'
        form.widgets.append(HtmlWidget(
                    htmltext('<%s%s>%s</%s>' % (enclosing_tag, extra_css_class,
                    label, enclosing_tag))))

    def add_to_view_form(self, *args):
        pass

    def fill_admin_form(self, form):
        form.add(TextWidget, 'label', title = _('Label'), value = self.label,
                required = True, cols = 70, rows = 3, render_br = False)
        form.add(StringWidget, 'extra_css_class', title = _('Extra class for CSS styling'),
                value = self.extra_css_class, size = 30)

    def get_admin_attributes(self):
        return Field.get_admin_attributes(self) + ['extra_css_class']


register_field_class(CommentField)



class StringField(WidgetField):
    key = 'string'
    description = N_('Text (line)')

    widget_class = WcsExtraStringWidget
    size = None
    extra_attributes = ['size']
    validation = None
    data_source = {}

    def perform_more_widget_changes(self, form, kwargs, edit=True):
        if self.data_source:
            if self.data_source.get('type') == 'jsonp':
                kwargs['url'] = self.data_source.get('value')
                self.widget_class = AutocompleteStringWidget

    def fill_admin_form(self, form):
        WidgetField.fill_admin_form(self, form)
        form.add(StringWidget, 'size', title = _('Line length'),
                value = self.size)
        form.add(RegexStringWidget, 'validation', title = _('Validation regex'),
                value = self.validation)
        form.add(data_sources.DataSourceSelectionWidget, 'data_source',
                 value=self.data_source,
                 title=_('External Data Source'),
                 hint=_('This will allow autocompletion from an external source.'),
                 required=False)

    def get_admin_attributes(self):
        return WidgetField.get_admin_attributes(self) + ['size', 'validation', 'data_source']

register_field_class(StringField)


class TextField(WidgetField):
    key = 'text'
    description = N_('Long Text')

    widget_class = TextWidget
    cols = None
    rows = None
    pre = False
    maxlength = None
    extra_attributes = ['cols', 'rows', 'maxlength']

    def fill_admin_form(self, form):
        WidgetField.fill_admin_form(self, form)
        form.add(StringWidget, 'cols', title = _('Line length'),
                value = self.cols)
        form.add(StringWidget, 'rows', title = _('Number of rows'),
                value = self.rows)
        form.add(StringWidget, 'maxlength',
                 title=_('Maximum number of characters'),
                 value = self.maxlength)
        form.add(CheckboxWidget, 'pre', title = _('Preformatted Text'),
                value = self.pre)

    def get_admin_attributes(self):
        return WidgetField.get_admin_attributes(self) + ['cols', 'rows', 'pre', 'maxlength']

    def get_view_value(self, value):
        if self.pre:
            return htmltext('<pre>') + value + htmltext('</pre>')
        else:
            try:
                return htmltext('<p>') + htmltext('\n').join(
                        [(x or htmltext('</p><p>')) for x in value.splitlines()]) + \
                    htmltext('</p>')
            except:
                return ''

    def get_view_short_value(self, value, max_len = 30):
        return ellipsize(str(value), max_len)

register_field_class(TextField)


class EmailField(WidgetField):
    key = 'email'
    description = N_('Email')

    widget_class = EmailWidget

    def get_view_value(self, value):
        return htmltext('<a href="mailto:%s">%s</a>') % (value, value)

    def get_rst_view_value(self, value, indent=''):
        return indent + value

register_field_class(EmailField)


class BoolField(WidgetField):
    key = 'bool'
    description = N_('Check Box (single choice)')

    widget_class = CheckboxWidget
    required = False
    anonymise = False

    def perform_more_widget_changes(self, form, kwargs, edit = True):
        if not edit:
            kwargs['disabled'] = 'disabled'
            value = get_request().get_field(self.field_key)
            form.add_hidden(self.field_key, value=str(value))
            if value and not value == 'False':
                self.field_key = 'f%sdisabled' % self.id
                get_request().form[self.field_key] = 'yes'
            self.field_key = 'f%sdisabled' % self.id

    def get_view_value(self, value):
        if value is True or value == 'True':
            return _('Yes')
        elif value is False or value == 'False':
            return _('No')
        else:
            return ''

    def convert_value_from_str(self, value):
        if value is None:
            return None
        if value in (True, 'True', _('Yes'), 'yes'):
            return True
        return False

    def convert_value_to_str(self, value):
        if value is True:
            return 'True'
        elif value is False:
            return 'False'
        return value

    def stats(self, values):
        no_records = len(values)
        r = TemplateIO(html = True)
        r += htmltext('<table class="stats">')
        r += htmltext('<thead><tr><th colspan="4">')
        r += self.label
        r += htmltext('</th></tr></thead>')
        options = (True, False)
        r += htmltext('<tbody>')
        for o in options:
            r += htmltext('<tr>')
            r += htmltext('<td class="label">')
            if o is True:
                r += _('Yes')
            else:
                r += _('No')
            r += htmltext('</td>')
            no = len([None for x in values if x.data.get(self.id) == str(o)])
            if no_records:
                r += htmltext('<td class="percent">')
                r += ' %.2f %%' % (100.*no/no_records)
                r += htmltext('</td>')
                r += htmltext('<td class="total">')
                r += '(%d/%d)' % (no, no_records)
                r += htmltext('</td>')
                image_len = 300.*no/no_records
                r += htmltext('<td>')
                if image_len:
                    r += htmltext('<img class="bar" src="%simages/bar.png" height="20px" alt="" '
                            'width="%dpx" />' % (get_publisher().get_root_url(), image_len))
                r += htmltext('</td>')
            else:
                r += htmltext('<td colspan="3"></td>')
                #'NaN'
            r += htmltext('</tr>')
        r += htmltext('</tbody>')
        r += htmltext('</table>')
        return r.getvalue()

register_field_class(BoolField)


class FileField(WidgetField):
    key = 'file'
    description = N_('File Upload')

    widget_class = FileWithPreviewWidget

    def get_view_value(self, value):
        return htmltext('<a download="%s" href="[download]?f=%s">%s</a>') % (
                value.base_filename, self.id, value)

    def get_csv_value(self, value, hint=None):
        if not value:
            value = ''
        if hint == 'ods' and value:
            return ['[download]files/%s/%s' % (self.id, value.base_filename)]
        return ['%s' % value]

    def get_json_value(self, value):
        return {'field_id': self.id, 'filename': value.base_filename}

    def perform_more_widget_changes(self, form, kwargs, edit = True):
        if not edit:
            value = get_request().get_field(self.field_key)
            if value and hasattr(value, 'token'):
                get_request().form[self.field_key + '$token'] = value.token

register_field_class(FileField)


class DateField(WidgetField):
    key = 'date'
    description = N_('Date')

    widget_class = DateWidget
    minimum_date = None
    maximum_date = None
    minimum_is_future = False
    date_in_the_past = False
    date_can_be_today = False
    extra_attributes = ['minimum_date', 'minimum_is_future', 'maximum_date',
            'date_in_the_past', 'date_can_be_today']

    def fill_admin_form(self, form):
        WidgetField.fill_admin_form(self, form)
        form.add(DateWidget, 'minimum_date', title = _('Minimum Date'), value = self.minimum_date)
        form.add(CheckboxWidget, 'minimum_is_future',
                title = _('Date must be in the future'), value = self.minimum_is_future,
                hint = _('This option is obviously not compatible with setting a minimum date'))
        form.add(DateWidget, 'maximum_date', title = _('Maximum Date'), value = self.maximum_date)
        form.add(CheckboxWidget, 'date_in_the_past',
                title = _('Date must be in the past'), value = self.date_in_the_past,
                hint = _('This option is obviously not compatible with setting a maximum date'))
        form.add(CheckboxWidget, 'date_can_be_today',
                title = _('Date can be present day'), value = self.date_can_be_today,
                hint = _('This option is only useful combined with one of the previous checkboxes.'))

    def get_admin_attributes(self):
        return WidgetField.get_admin_attributes(self) + [
                'minimum_date', 'minimum_is_future', 'maximum_date',
                'date_in_the_past', 'date_can_be_today']

    def convert_value_from_str(self, value):
        return time.strptime(value, date_format())

    def convert_value_to_str(self, value):
        if value is None:
            return ''
        try:
            return strftime(date_format(), value)
        except TypeError:
            return ''

    def add_to_form(self, form, value=None):
        if value and type(value) is not str:
            value = self.convert_value_to_str(value)
        WidgetField.add_to_form(self, form, value=value)

    def add_to_view_form(self, form, value = None):
        value = localstrftime(value)
        return WidgetField.add_to_view_form(self, form, value = value)

    def get_view_value(self, value):
        try:
            return strftime(misc.date_format(), value)
        except TypeError:
            return value

register_field_class(DateField)


def item_items_stats(field, values):
    no_records = len(values)
    r = TemplateIO(html = True)
    r += htmltext('<table class="stats">')
    r += htmltext('<thead><tr><th colspan="4">')
    r += field.label
    r += htmltext('</th></tr></thead>')
    options = field.items or []
    r += htmltext('<tbody>')
    for o in options:
        r += htmltext('<tr>')
        r += htmltext('<td class="label">')
        r += o
        r += htmltext('</td>')
        if field.type == 'item':
            no = len([None for x in values if x.data.get(field.id) == o])
        else:
            no = len([None for x in values if o in (x.data.get(field.id) or [])])
        if no_records:
            r += htmltext('<td class="percent">')
            r += ' %.2f %%' % (100.*no/no_records)
            r += htmltext('</td>')
            r += htmltext('<td class="total">')
            r += '(%d/%d)' % (no, no_records)
            r += htmltext('</td>')
            image_len = 300.*no/no_records
            r += htmltext('<td>')
            if image_len:
                r += htmltext('<img class="bar" src="%simages/bar.png" height="20px" alt="" '
                        'width="%dpx" />' % (get_publisher().get_root_url(), image_len))
            r += htmltext('</td>')
        else:
            r += htmltext('<td colspan="3"></td>')
            #'NaN'
        r += htmltext('</tr>')
    r += htmltext('</tbody>')
    r += htmltext('</table>')
    return r.getvalue()


class ItemField(WidgetField):
    key = 'item'
    description = N_('List')

    items = []
    show_as_radio = False
    anonymise = False
    widget_class = SingleSelectHintWidget
    data_source = {}

    def __init__(self, **kwargs):
        self.items = []
        WidgetField.__init__(self, **kwargs)

    def perform_more_widget_changes(self, form, kwargs, edit = True):
        if self.data_source:
            if self.data_source.get('type') == 'jsonp':
                kwargs['url'] = self.data_source.get('value')
                self.widget_class = JsonpSingleSelectWidget
            else:
                kwargs['options'] = data_sources.get_items(self.data_source)
                if kwargs['options'] and not self.required:
                    if type(kwargs['options'][0]) is str:
                        kwargs['options'][:0] = [None]
                    elif len(kwargs['options'][0]) == 2:
                        kwargs['options'][:0] = [(None, '---')]
                    elif len(kwargs['options'][0]) == 3:
                        kwargs['options'][:0] = [(None, '---', None)]
        elif self.items:
            kwargs['options'] = self.items
        if not kwargs.get('options'):
            kwargs['options'] = [(None, '---')]
        if self.show_as_radio:
            self.widget_class = RadiobuttonsWidget
            if type(kwargs['options'][0]) is str:
                first_item = kwargs['options'][0]
            else:
                first_item = kwargs['options'][0][1]
            if len(kwargs['options']) > 3 or len(first_item) > 10:
                # TODO: absence/presence of delimitor should be an option
                kwargs['delim'] = htmltext('<br />')

    def get_display_value(self, value):
        real_value = value
        label_value = str(value or '')
        kwargs = {}
        self.perform_more_widget_changes(form, kwargs, False)
        if self.data_source and self.data_source.get('type') == 'jsonp':
            if not get_session().jsonp_display_values:
                get_session().jsonp_display_values = {}
            label_value = get_session().jsonp_display_values.get(
                            '%s_%s' % (self.data_source.get('value'), value))
        elif type(kwargs['options'][0]) in (tuple, list):
            if len(kwargs['options'][0]) == 2:
                for key, value in kwargs['options']:
                    if str(key) == str(real_value):
                        label_value = value
                        break
            elif len(kwargs['options'][0]) == 3:
                for key, value, key2 in kwargs['options']:
                    if str(key) == str(real_value):
                        label_value = value
                        break
        return label_value

    def add_to_view_form(self, form, value = None):
        real_value = value
        label_value = self.get_display_value(value)
        self.field_key = 'f%s' % self.id

        get_request().form[self.field_key + '_label'] = label_value # :/
        form.add(StringWidget, self.field_key + '_label', title = self.label,
                 value=label_value, readonly = 'readonly', size=len(label_value or ''))
        form.add(HiddenWidget, self.field_key, value=real_value)

    def store_display_value(self, data, field_id):
        value = data.get(field_id)
        kwargs = {}
        self.perform_more_widget_changes(None, kwargs, False)
        if self.data_source and self.data_source.get('type') == 'jsonp':
            display_value = get_request().form.get('f%s_display' % field_id)
            if display_value is None:
                if not get_session().jsonp_display_values:
                    get_session().jsonp_display_values = {}
                display_value = get_session().jsonp_display_values.get(
                                '%s_%s' % (self.data_source.get('value'), value))
            else:
                if not get_session().jsonp_display_values:
                    get_session().jsonp_display_values = {}
                get_session().jsonp_display_values[
                        '%s_%s' % (self.data_source.get('value'), value)] = display_value
            return display_value
        elif type(kwargs['options'][0]) in (tuple, list):
            if len(kwargs['options'][0]) == 2:
                for key, option_value in kwargs['options']:
                    if str(key) == str(value):
                        return option_value
            elif len(kwargs['options'][0]) == 3:
                for key, option_value, key_repeat in kwargs['options']:
                    if str(key) == str(value):
                        return option_value
        return str(value)

    def fill_admin_form(self, form):
        WidgetField.fill_admin_form(self, form)
        form.add(CheckboxWidget, 'show_as_radio', title = _('Show as radio buttons'),
                value = self.show_as_radio)
        form.add(WidgetList, 'items', title = _('Items'), element_type = StringWidget,
                value = self.items, required = False,
                element_kwargs = {'render_br': False, 'size': 50},
                add_element_label = _('Add item'))
        form.add(data_sources.DataSourceSelectionWidget, 'data_source',
                 value=self.data_source,
                 title=_('External Data Source'),
                 hint=_('This will get the available items from an external source.'),
                 required=False)

    def get_admin_attributes(self):
        return WidgetField.get_admin_attributes(self) + ['items', 'show_as_radio', 'data_source']

    def check_admin_form(self, form):
        items = form.get_widget('items').parse()
        d = {}
        for v in (items or []):
            if v in d:
                form.set_error('items', _('Duplicated Items'))
                return
            d[v] = None

    def stats(self, values):
        return item_items_stats(self, values)

    def feed_session(self, value, display_value):
        if self.data_source and self.data_source.get('type') == 'jsonp':
            if not get_session().jsonp_display_values:
                get_session().jsonp_display_values = {}
            get_session().jsonp_display_values[
                    '%s_%s' % (self.data_source.get('value'), value)] = display_value

register_field_class(ItemField)


class ItemsField(WidgetField):
    key = 'items'
    description = N_('Multiple choice list')

    items = []
    max_choices = 0
    data_source = {}

    widget_class = CheckboxesWidget

    _cached_data_source = None

    def __init__(self, **kwargs):
        self.items = []
        WidgetField.__init__(self, **kwargs)

    def get_items(self):
        if self.data_source:
            if self._cached_data_source:
                return self._cached_data_source
            self._cached_data_source = data_sources.get_items(self.data_source)
            return self._cached_data_source
        elif self.items:
            return self.items
        else:
            return []

    def perform_more_widget_changes(self, form, kwargs, edit = True):
        kwargs['elements'] = self.get_items()
        kwargs['max_choices'] = self.max_choices

        if len(self.get_items()) > 3:
            kwargs['inline'] = False

    def fill_admin_form(self, form):
        WidgetField.fill_admin_form(self, form)
        form.add(WidgetList, 'items', title = _('Items'), element_type = StringWidget,
                value = self.items, required = False,
                element_kwargs = {'render_br': False, 'size': 50},
                add_element_label = _('Add item'))
        form.add(IntWidget, 'max_choices', title = _('Maximum number of choices'),
                value = self.max_choices, required = False, size = 4)
        form.add(data_sources.DataSourceSelectionWidget, 'data_source',
                 value=self.data_source,
                 allow_jsonp=False,
                 title=_('External Data Source'),
                 hint=_('This will get the available items from an external source.'),
                 required=False)

    def get_admin_attributes(self):
        return WidgetField.get_admin_attributes(self) + ['items',
                'max_choices', 'data_source']

    def check_admin_form(self, form):
        items = form.get_widget('items').parse()
        d = {}
        for v in (items or []):
            if v in d:
                form.set_error('items', _('Duplicated Items'))
                return
            d[v] = None

    def convert_value_to_str(self, value):
        return value

    def get_view_value(self, value):
        if type(value) is str: # == display_value
            return value
        if value:
            try:
                return ', '.join([(x) for x in value])
            except TypeError:
                pass
        return ''

    def stats(self, values):
        return item_items_stats(self, values)

    def get_csv_heading(self):
        labels = [self.label]
        if self.max_choices:
            labels.extend(['']*(self.max_choices-1))
        elif len(self.get_items()):
            labels.extend(['']*(len(self.get_items())-1))
        return labels

    def get_csv_value(self, value, hint=None):
        values = []
        for one_value in value:
            values.append(one_value)
        if self.max_choices:
            nb_columns = self.max_choices
        elif len(self.get_items()):
            nb_columns = len(self.get_items())
        else:
            nb_columns = 1

        if len(values) > nb_columns:
            # this would happen if max_choices is set after forms were already
            # filled with more values
            values = values[:nb_columns]
        elif len(values) < nb_columns:
            values.extend(['']*(nb_columns-len(values)))
        return values

    def store_display_value(self, data, field_id):
        kwargs = {}
        options = self.get_items()
        choices = []
        for choice in data.get(field_id) or []:
            if type(options[0]) is str:
                choices.append(choice)
            elif type(options[0]) in (tuple, list):
                if len(options[0]) == 2:
                    for key, option_value in options:
                        if str(key) == str(choice):
                            choices.append(option_value)
                            break
                elif len(options[0]) == 3:
                    for key, option_value, key_repeat in options:
                        if str(key) == str(choice):
                            choices.append(option_value)
                            break
        return ', '.join(choices)
register_field_class(ItemsField)


class PageField(Field):
    key = 'page'
    description = N_('New Page')

    condition = None

    def fill_admin_form(self, form):
        form.add(StringWidget, 'label', title = _('Label'), value = self.label,
                required = True, size = 50)
        form.add(StringWidget, 'condition', title = _('Condition'), value = self.condition,
                required = False, size = 50)

    def get_admin_attributes(self):
        return Field.get_admin_attributes(self) + ['condition']

    def add_to_view_form(self, *args):
        pass

    def is_visible(self, dict, formdef):
        if dict is None:
            return True

        if not self.condition:
            return True

        from formdata import get_dict_with_varnames
        data = get_dict_with_varnames(formdef.fields, dict)

        try:
            if eval(self.condition, get_publisher().get_global_eval_dict(), data):
                return True
        except Exception, e:
            get_logger().warn('failed to evaluate condition "%s" (%r)' % (self.condition, e))
            return True

        return False

register_field_class(PageField)


class TableField(WidgetField):
    key = 'table'
    description = N_('Table')
    in_listing = False # no way to represent data in a single cell

    rows = None
    columns = None

    widget_class = TableWidget

    def __init__(self, **kwargs):
        self.rows = []
        self.columns = []
        WidgetField.__init__(self, **kwargs)

    def perform_more_widget_changes(self, form, kwargs, edit = True):
        kwargs['rows'] = self.rows
        kwargs['columns'] = self.columns

    def fill_admin_form(self, form):
        WidgetField.fill_admin_form(self, form)
        try:
            form.remove('in_listing')
        except KeyError: # perhaps it was already removed
            pass
        try:
            form.remove('prefill')
        except KeyError: # perhaps it was already removed
            pass
        form.add(WidgetList, 'rows', title = _('Rows'), element_type = StringWidget,
                value = self.rows, required = True,
                element_kwargs = {'render_br': False, 'size': 50},
                add_element_label = _('Add row'))
        form.add(WidgetList, 'columns', title = _('Columns'), element_type = StringWidget,
                value = self.columns, required = True,
                element_kwargs = {'render_br': False, 'size': 50},
                add_element_label = _('Add column'))

    def get_admin_attributes(self):
        t = WidgetField.get_admin_attributes(self) + ['rows', 'columns']
        try:
            t.remove('in_listing')
        except ValueError:
            pass
        try:
            t.remove('prefill')
        except ValueError:
            pass
        return t

    def get_view_value(self, value):
        r = TemplateIO(html=True)
        r += htmltext('<table><thead><tr><td></td>')
        for column in self.columns:
            r += htmltext('<th>%s</th>') % column
        r += htmltext('</tr></thead><tbody>')
        for i, row in enumerate(self.rows):
            r += htmltext('<tr><th>%s</th>') % row
            for j, column in enumerate(self.columns):
                r += htmltext('<td>')
                if value:
                    try:
                        r += value[i][j]
                    except IndexError:
                        pass
                r += htmltext('</td>')
            r += htmltext('</tr>')
        r += htmltext('</tbody></table>')

        return r.getvalue()

    def get_rst_view_value(self, value, indent=''):
        if not value:
            return indent
        r = []
        max_width = 0
        for column in self.columns:
            max_width = max(max_width, len(column))

        for i, row in enumerate(value):
            value[i] = [x or '' for x in row]

        for i, row in enumerate(self.rows):
            max_width = max(max_width, len(row))
            for j, column in enumerate(self.columns):
                max_width = max(max_width, len(value[i][j]))

        r.append(' '.join(['='*max_width]*(len(self.columns)+1)))
        r.append(' '.join([column.center(max_width) for column in ['/']+self.columns]))
        r.append(' '.join(['='*max_width]*(len(self.columns)+1)))
        for i, row in enumerate(self.rows):
            r.append(' '.join([cell.center(max_width) for cell in [row] +
                    [value[i][x] for x in range(len(self.columns))]]))
        r.append(' '.join(['='*max_width]*(len(self.columns)+1)))
        return '\n'.join([indent + x for x in r])

    def get_csv_heading(self):
        if not self.columns:
            return [self.label]
        labels = []
        for col in self.columns:
            for row in self.rows:
                t = '%s / %s' % (col, row)
                if len(labels) == 0:
                    labels.append('%s - %s' % (self.label, t))
                else:
                    labels.append(t)
        return labels

    def get_csv_value(self, value, hint=None):
        if not self.columns:
            return ['']
        values = []
        for i in range(len(self.columns)):
            for j in range(len(self.rows)):
                try:
                    values.append(value[j][i])
                except IndexError:
                    values.append('')
        return values

register_field_class(TableField)


class TableSelectField(TableField):
    key = 'table-select'
    description = N_('Table of Lists')
    in_listing = True

    items = None

    widget_class = SingleSelectTableWidget

    def __init__(self, **kwargs):
        self.items = []
        TableField.__init__(self, **kwargs)

    def fill_admin_form(self, form):
        TableField.fill_admin_form(self, form)
        form.add(WidgetList, 'items', title=_('Items'), element_type=StringWidget,
                value=self.items, required=True,
                element_kwargs = {'render_br': False, 'size': 50},
                add_element_label = _('Add item'))

    def perform_more_widget_changes(self, form, kwargs, edit=True):
        TableField.perform_more_widget_changes(self, form, kwargs, edit=edit)
        if edit:
            kwargs['options'] = self.items or [(None, '---')]
        else:
            self.widget_class = TableWidget

    def get_admin_attributes(self):
        return TableField.get_admin_attributes(self) + ['items']

    def check_admin_form(self, form):
        items = form.get_widget('items').parse()
        d = {}
        for v in (items or []):
            if v in d:
                form.set_error('items', _('Duplicated Items'))
                return
            d[v] = None

register_field_class(TableSelectField)


class TableRowsField(WidgetField):
    key = 'tablerows'
    description = N_('Table with rows')
    in_listing = False # no way to represent data in a single cell

    total_row = True
    columns = None

    widget_class = TableListRowsWidget

    def __init__(self, **kwargs):
        self.columns = []
        WidgetField.__init__(self, **kwargs)

    def perform_more_widget_changes(self, form, kwargs, edit = True):
        kwargs['columns'] = self.columns
        kwargs['add_element_label'] = _('Add row')

    def fill_admin_form(self, form):
        WidgetField.fill_admin_form(self, form)
        try:
            form.remove('in_listing')
        except KeyError: # perhaps it was already removed
            pass
        try:
            form.remove('prefill')
        except KeyError: # perhaps it was already removed
            pass
        form.add(WidgetList, 'columns', title = _('Columns'), element_type = StringWidget,
                value = self.columns, required = True,
                element_kwargs = {'render_br': False, 'size': 50},
                add_element_label = _('Add column'))
        form.add(CheckboxWidget, 'total_row', title = _('Total Row'),
                value = self.total_row)

    def get_admin_attributes(self):
        t = WidgetField.get_admin_attributes(self) + ['columns', 'total_row']
        try:
            t.remove('in_listing')
        except ValueError:
            pass
        try:
            t.remove('prefill')
        except ValueError:
            pass
        return t

    def get_view_value(self, value):
        r = TemplateIO(html=True)
        r += htmltext('<table><thead><tr>')
        for column in self.columns:
            r += htmltext('<th>%s</th>') % column
        r += htmltext('</tr></thead><tbody>')
        for i, row in enumerate(value):
            r += htmltext('<tr>')
            for j, column in enumerate(self.columns):
                r += htmltext('<td>')
                if value:
                    try:
                        r += row[j]
                    except IndexError:
                        pass
                r += htmltext('</td>')
            r += htmltext('</tr>')
        r += htmltext('</tbody>')

        if self.total_row:
            sums_row = []
            for j, column in enumerate(self.columns):
                sum_column = 0
                for row_value in value:
                    try:
                        cell_value = row_value[j]
                    except IndexError:
                        continue
                    if cell_value in (None, ''):
                        continue
                    try:
                        sum_column += float(cell_value)
                    except ValueError:
                        sums_row.append(None)
                        break
                else:
                    sums_row.append(sum_column)
            if [x for x in sums_row if x is not None]:
                r += htmltext('<tfoot><tr>')
                for sum_column in sums_row:
                    if sum_column is None:
                        r += htmltext('<td></td>')
                    else:
                        r += htmltext('<td>%.2f</td>' % sum_column)
                r += htmltext('</tr></tfoot>')

        r += htmltext('</table>')

        return r.getvalue()

    def get_rst_view_value(self, value, indent=''):
        if not value:
            return indent
        r = []
        max_width = 0
        for column in self.columns:
            max_width = max(max_width, len(column))

        for i, row in enumerate(value):
            value[i] = [x or '' for x in row]

        for i, row_value in enumerate(value):
            for j, column in enumerate(self.columns):
                max_width = max(max_width, len(row_value[j]))

        r.append(' '.join(['='*max_width]*(len(self.columns))))
        r.append(' '.join([column.center(max_width) for column in self.columns]))
        r.append(' '.join(['='*max_width]*(len(self.columns))))
        for i, row_value in enumerate(value):
            r.append(' '.join([cell.center(max_width) for cell in
                    [row_value[x] for x in range(len(self.columns))]]))
        r.append(' '.join(['='*max_width]*(len(self.columns))))
        return '\n'.join([indent + x for x in r])

    def get_csv_value(self, element, hint=None):
        return [_('unimplemented')] # XXX


register_field_class(TableRowsField)


class RankedItemsField(WidgetField):
    key = 'ranked-items'
    description = N_('Ranked Items')

    items = []
    randomize_items = False
    widget_class = RankedItemsWidget
    anonymise = False

    def perform_more_widget_changes(self, form, kwargs, edit = True):
        kwargs['elements'] = self.items or []
        kwargs['randomize_items'] = self.randomize_items

    def fill_admin_form(self, form):
        WidgetField.fill_admin_form(self, form)
        form.add(WidgetList, 'items', title = _('Items'), element_type = StringWidget,
                value = self.items, required = True,
                element_kwargs = {'render_br': False, 'size': 50},
                add_element_label = _('Add item'))
        form.add(CheckboxWidget, 'randomize_items', title = _('Randomize Items'),
                value = self.randomize_items)

    def get_admin_attributes(self):
        return WidgetField.get_admin_attributes(self) + ['items', 'randomize_items']

    def get_view_value(self, value):
        r = TemplateIO(html=True)
        r += htmltext('<ul>')
        items = value.items()
        items.sort(lambda x,y: cmp(x[1], y[1]))
        counter = 0
        last_it = None
        for it in items:
            if it[1] is not None:
                if last_it != it[1]:
                    counter += 1
                    last_it = it[1]
                r += htmltext('<li>%s: %s</li>') % (counter, it[0])
        r += htmltext('</ul>')
        return r.getvalue()

    def get_rst_view_value(self, value, indent=''):
        items = value.items()
        items.sort(lambda x,y: cmp(x[1], y[1]))
        counter = 0
        last_it = None
        values = []
        for it in items:
            if it[1] is not None:
                if last_it != it[1]:
                    counter += 1
                    last_it = it[1]
                values.append('%s: %s' % (counter, it[0]))
        return indent + ' / '.join(values)

    def get_csv_heading(self):
        if not self.items:
            return [self.label]
        return [self.label] + ['']*(len(self.items)-1)

    def get_csv_value(self, value, hint=None):
        if not self.items:
            return ['']
        if type(value) is not dict:
            value = {}
        items = value.items()
        items.sort(lambda x,y: cmp(x[1], y[1]))
        ranked = [x[0] for x in items if x[1] is not None]
        return ranked + ['' for x in range(len(self.items)-len(ranked))]


register_field_class(RankedItemsField)


def get_field_class_by_type(type):
    for k in field_classes:
        if k.key == type:
            return k
    raise KeyError()

def get_field_types():
    return field_types

