# w.c.s. - web application for online forms
# Copyright (C) 2005-2010  Entr'ouvert
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

import base64
import json

from quixote import get_session, get_publisher
import quixote.http_request
from quixote.errors import RequestError

from http_response import HTTPResponse

class HTTPRequest(quixote.http_request.HTTPRequest):
    def __init__(self, *args, **kwargs):
        quixote.http_request.HTTPRequest.__init__(self, *args, **kwargs)
        self.response = HTTPResponse()
        self.charset = get_publisher().site_charset

    _user = () # use empty tuple instead of None as None is a "valid" user value
    def get_user(self):
        if self._user != ():
            return self._user

        auth_header = self.get_header('Authorization', '')
        if auth_header.startswith('Basic '):
            auth_header = auth_header.split(' ', 1)[1]
            username, password = base64.decodestring(auth_header).split(':', 1)
            from qommon.ident.password_accounts import PasswordAccount
            try:
                self._user = PasswordAccount.get_with_credentials(username, password)
            except KeyError:
                self._user = None
            return

        try:
            self._user = get_session().get_user()
        except AttributeError:
            self._user = None
        return self._user
    user = property(get_user)

    def get_server(self, clean = True):
        server_name = quixote.http_request.HTTPRequest.get_server(self)
        if clean and server_name.lower()[:7] in ('iframe.', 'iframe-'):
            server_name = server_name[7:]
        return server_name

    def get_local_url(self, n=0):
        '''Return the local part of the URL, query string included'''
        query = self.get_query()
        if query:
            return self.get_path(n) + '?' + query
        else:
            return self.get_path(n)

    def get_frontoffice_url(self, n=0):
        return get_publisher().get_frontoffice_url(without_script_name=True) + \
                self.get_local_url(n)

    def get_substitution_variables(self):
        # environment variables APPNAME_* are exported to env_*
        prefix = get_publisher().APP_NAME.lower() + '_'
        variables = {}
        for k, v in self.environ.items():
            if k.lower().startswith(prefix):
                variables['env_' + k.lower()[len(prefix):]] = v
        return variables

    def dump(self):
        # straight copy of HTTPRequest.dump(), sole modification is that the
        # values are printed as %r, not %s
        result=[]
        row='%-15s %r'

        result.append("Form:")
        L = self.form.items() ; L.sort()
        for k,v in L:
            result.append(row % (k,v))

        result.append("")
        result.append("Cookies:")
        L = self.cookies.items() ; L.sort()
        for k,v in L:
            result.append(row % (k,v))


        result.append("")
        result.append("Environment:")
        L = self.environ.items() ; L.sort()
        for k,v in L:
            result.append(row % (k,v))
        return "\n".join(result)

    def process_inputs(self):
        quixote.http_request.HTTPRequest.process_inputs(self)
        ctype = self.environ.get("CONTENT_TYPE")
        if ctype == 'application/json':
            length = int(self.environ.get('CONTENT_LENGTH') or '0')
            payload = self.stdin.read(length)
            try:
                self.json = json.loads(payload)
            except ValueError, e:
                raise RequestError('invalid json payload (%s)' % str(e))
        # Make sure request.form doesn't contain unicode strings, converting
        # them all to strings encoded in the site charset.
        for k, v in self.form.items():
            if type(v) is unicode:
                self.form[k] = v.encode(self.charset)

    def is_json(self):
        if self.get_header('Accept') == 'application/json':
            return True
        if self.get_query() == 'json':
            return True
        if self.form and self.form.get('format') == 'json':
            return True
        return False
