# w.c.s. - web application for online forms
# Copyright (C) 2005-2010  Entr'ouvert
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

import x509utils
import os.path

def bool2xs(boolean):
    '''Convert a boolean value to XSchema boolean representation'''
    if boolean is True:
        return 'true'
    if boolean is False:
        return 'false'
    raise TypeError()

class Metadata(object):
    __endpoints = {
            'slo' : 'singleLogout',
            'fedterm' : 'federationTermination',
            'rni' : 'registerNameIdentifier',
            'ac' : 'assertionConsumer',
            'se' : 'soapEndpoint' }

    def __init__(self, publisher, provider_id, config):
        self.publisher = publisher
        self.provider_id = provider_id
        self.config = config

    def get_key_descriptor(self, keytype, key):
        '''Format key as an XML Dsig KeyNode content'''
        if keytype:
            prologue = '    <KeyDescriptor use="%s">' % keytype
        else:
            prologue = '    <KeyDescriptor>'
        if key and 'CERTIF' in key:
            naked = x509utils.decapsulate_pem_file(key)
            return prologue + '''
      <ds:KeyInfo xmlns:ds="http://www.w3.org/2000/09/xmldsig#">
        <ds:X509Data><ds:X509Certificate>%s</ds:X509Certificate></ds:X509Data>
      </ds:KeyInfo>
    </KeyDescriptor>
''' % naked
        # FIXME: generate proper RSAKeyValue, but wait for support in Lasso
        elif key and 'KEY' in key:
            naked = x509utils.get_xmldsig_rsa_key_value(key)
            return prologue  + '''
      <ds:KeyInfo xmlns:ds="http://www.w3.org/2000/09/xmldsig#">
        %s
      </ds:KeyInfo>
    </KeyDescriptor>
''' % naked
        else:
            return ''

    def get_new_or_old_keys(self, signing_pem_key, encryption_pem_key):
        '''Return new or earlier version of PEM keys'''
        dir = self.publisher.app_dir
        if not signing_pem_key and self.config.get('publickey'):
            signing_pem_key = file(os.path.join(dir, 'public-key.pem')).read()
        if not encryption_pem_key and self.config.get('encryption_publickey'):
            encryption_pem_key = file(os.path.join(dir, 'encryption-public-key.pem')).read()
        return (signing_pem_key, encryption_pem_key)

    def get_key_descriptors(self, signing_pem_key, encryption_pem_key):
        signing_pem_key, encryption_pem_key = \
            self.get_new_or_old_keys(signing_pem_key, encryption_pem_key)
        sp_key = {
                'signing': self.get_key_descriptor('signing', signing_pem_key),
                'encryption': self.get_key_descriptor('encryption', encryption_pem_key) }
        if not sp_key['signing'] and sp_key['encryption']:
            sp_key = {
                'signing': '',
                'encryption': self.get_key_descriptor('', encryption_pem_key) }
        if sp_key['signing'] and not sp_key['encryption']:
            sp_key = {
                'signing': self.get_key_descriptor('', signing_pem_key),
                'encryption': ''
            }
        return sp_key

    def get_spsso_descriptor(self, signing_pem_key, encryption_pem_key, endpoints):
        authnrequestsigned = bool2xs(self.config.get('authn-request-signed', True))
        sp_key = self.get_key_descriptors(signing_pem_key, encryption_pem_key)
        prologue = '''  <SPDescriptor protocolSupportEnumeration="urn:liberty:iff:2003-08">
'''
        config = dict(self.config)
        config.update({ 'authn-requests-signed': authnrequestsigned })
        config.update(self.__endpoints)
        config.update(endpoints)

        return prologue + sp_key['signing'] + sp_key['encryption'] +   '''
    <SoapEndpoint>%(base_url)s/%(se)s</SoapEndpoint>

    <SingleLogoutServiceURL>%(base_url)s/%(slo)s</SingleLogoutServiceURL>
    <SingleLogoutServiceReturnURL>%(base_url)s/%(slo)sReturn</SingleLogoutServiceReturnURL>

    <FederationTerminationServiceURL>%(base_url)s/%(fedterm)s</FederationTerminationServiceURL>
    <FederationTerminationServiceReturnURL>%(base_url)s/%(fedterm)sReturn</FederationTerminationServiceReturnURL>
    <FederationTerminationNotificationProtocolProfile>http://projectliberty.org/profiles/fedterm-idp-soap</FederationTerminationNotificationProtocolProfile>
    <FederationTerminationNotificationProtocolProfile>http://projectliberty.org/profiles/fedterm-idp-http</FederationTerminationNotificationProtocolProfile>
    <FederationTerminationNotificationProtocolProfile>http://projectliberty.org/profiles/fedterm-sp-soap</FederationTerminationNotificationProtocolProfile>
    <FederationTerminationNotificationProtocolProfile>http://projectliberty.org/profiles/fedterm-sp-http</FederationTerminationNotificationProtocolProfile>

    <SingleLogoutProtocolProfile>http://projectliberty.org/profiles/slo-idp-soap</SingleLogoutProtocolProfile>
    <SingleLogoutProtocolProfile>http://projectliberty.org/profiles/slo-idp-http</SingleLogoutProtocolProfile>
    <SingleLogoutProtocolProfile>http://projectliberty.org/profiles/slo-sp-soap</SingleLogoutProtocolProfile>
    <SingleLogoutProtocolProfile>http://projectliberty.org/profiles/slo-sp-http</SingleLogoutProtocolProfile>

    <RegisterNameIdentifierServiceURL>%(base_url)s/%(rni)s</RegisterNameIdentifierServiceURL>
    <RegisterNameIdentifierServiceReturnURL>%(base_url)s/%(rni)sReturn</RegisterNameIdentifierServiceReturnURL>
    <RegisterNameIdentifierProtocolProfile>http://projectliberty.org/profiles/rni-idp-soap</RegisterNameIdentifierProtocolProfile>
    <RegisterNameIdentifierProtocolProfile>http://projectliberty.org/profiles/rni-idp-http</RegisterNameIdentifierProtocolProfile>
    <RegisterNameIdentifierProtocolProfile>http://projectliberty.org/profiles/rni-sp-soap</RegisterNameIdentifierProtocolProfile>
    <RegisterNameIdentifierProtocolProfile>http://projectliberty.org/profiles/rni-sp-http</RegisterNameIdentifierProtocolProfile>

    <AuthnRequestsSigned>%(authn-requests-signed)s</AuthnRequestsSigned>

    <AssertionConsumerServiceURL id="AssertionConsumerServiceURL1" isDefault="true">%(base_url)s/%(ac)s</AssertionConsumerServiceURL>


</SPDescriptor>''' % config

    def get_idpsso_descriptor(self, signing_pem_key, encryption_pem_key):
        idp_key = self.get_key_descriptors(signing_pem_key, encryption_pem_key)
        idp_head = """  <IDPDescriptor protocolSupportEnumeration="urn:liberty:iff:2003-08">
"""
        idp_body = """
    <SoapEndpoint>%(base_soap_url)s/soapEndpoint</SoapEndpoint>

    <SingleLogoutServiceURL>%(base_url)s/singleLogout</SingleLogoutServiceURL>
    <SingleLogoutServiceReturnURL>%(base_url)s/singleLogoutReturn</SingleLogoutServiceReturnURL>

    <FederationTerminationServiceURL>%(base_url)s/federationTermination</FederationTerminationServiceURL>
    <FederationTerminationServiceReturnURL>%(base_url)s/federationTerminationReturn</FederationTerminationServiceReturnURL>
    <FederationTerminationNotificationProtocolProfile>http://projectliberty.org/profiles/fedterm-idp-soap</FederationTerminationNotificationProtocolProfile>
    <FederationTerminationNotificationProtocolProfile>http://projectliberty.org/profiles/fedterm-idp-http</FederationTerminationNotificationProtocolProfile>
    <FederationTerminationNotificationProtocolProfile>http://projectliberty.org/profiles/fedterm-sp-soap</FederationTerminationNotificationProtocolProfile>
    <FederationTerminationNotificationProtocolProfile>http://projectliberty.org/profiles/fedterm-sp-http</FederationTerminationNotificationProtocolProfile>

    <SingleLogoutProtocolProfile>http://projectliberty.org/profiles/slo-idp-soap</SingleLogoutProtocolProfile>
    <SingleLogoutProtocolProfile>http://projectliberty.org/profiles/slo-idp-http</SingleLogoutProtocolProfile>
    <SingleLogoutProtocolProfile>http://projectliberty.org/profiles/slo-sp-soap</SingleLogoutProtocolProfile>
    <SingleLogoutProtocolProfile>http://projectliberty.org/profiles/slo-sp-http</SingleLogoutProtocolProfile>

    <RegisterNameIdentifierProtocolProfile>http://projectliberty.org/profiles/rni-idp-soap</RegisterNameIdentifierProtocolProfile>
    <RegisterNameIdentifierProtocolProfile>http://projectliberty.org/profiles/rni-idp-http</RegisterNameIdentifierProtocolProfile>
    <RegisterNameIdentifierProtocolProfile>http://projectliberty.org/profiles/rni-sp-soap</RegisterNameIdentifierProtocolProfile>
    <RegisterNameIdentifierProtocolProfile>http://projectliberty.org/profiles/rni-sp-http</RegisterNameIdentifierProtocolProfile>
    <RegisterNameIdentifierServiceURL>%(base_url)s/registerNameIdentifier</RegisterNameIdentifierServiceURL>
    <RegisterNameIdentifierServiceReturnURL>%(base_url)s/registerNameIdentifierReturn</RegisterNameIdentifierServiceReturnURL>

    <SingleSignOnServiceURL>%(base_url)s/singleSignOn</SingleSignOnServiceURL>
    <SingleSignOnProtocolProfile>http://projectliberty.org/profiles/brws-art</SingleSignOnProtocolProfile>
    <SingleSignOnProtocolProfile>http://projectliberty.org/profiles/brws-post</SingleSignOnProtocolProfile>

  </IDPDescriptor>""" % self.config

        return idp_head + idp_key['signing'] + idp_key['encryption'] + idp_body

    def get_metadata(self, signing_pem_key = '', encryption_pem_key = '', do_idp = False, do_sp = False, endpoints = {}):
        prologue = '''<?xml version="1.0"?>
<EntityDescriptor
    providerID="%s"
    xmlns="urn:liberty:metadata:2003-08">\n''' % self.provider_id

        sp_descriptor = ''
        if do_sp:
            sp_descriptor = self.get_spsso_descriptor(signing_pem_key, encryption_pem_key, endpoints)

        idp_descriptor = ''
        if do_idp:
            idp_descriptor = self.get_idpsso_descriptor(signing_pem_key, encryption_pem_key)

        orga = ''
        if self.config.get('organization_name'):
            orga = '''<Organization>
   <OrganizationName xml:lang="en">%s</OrganizationName>
</Organization>''' % self.publisher.sitecharset2utf8(self.config['organization_name'])

        epilogue = '</EntityDescriptor>'

        return '\n'.join([prologue, sp_descriptor, idp_descriptor, orga, epilogue])

if __name__ == '__main__':
    print 'Testing metadata generation'
    class NoPublisher(object):
        pass
    publisher = NoPublisher()
    publisher.app_dir = '/dir'
    pkey, _ = x509utils.generate_rsa_keypair()
    meta = Metadata(publisher = publisher, config = { 'base_url': 'base_ulr', 'base_soap_url': 'base_soap_url' }, provider_id = 'provider_id_1')
    assert meta != None
    content = meta.get_saml2_metadata(pkey,'',True,True)
    assert isinstance(content, str) and content != ''
    print content

